#!/usr/bin/env python
# coding=utf-8

"""
todotxt_machine
~~~~~~~~~~~~~~~

An interactive terminal based todo.txt file editor with an interface similar to mutt

"""

__name__ = "todotxt_machine"
__title__ = "todotxt_machine"
__author__ = "Anthony DiGirolamo"
__license__ = "GPL3"
__version__ = (2, 0, 5)

version = "%s.%s.%s" % __version__

__all__ = ["cli", "todo", "colorscheme", "urwid_ui", "terminal_operations"]



#!/usr/bin/env python
# coding=utf-8
import re
import random
import urwid
from datetime import date

from todotxt_machine.terminal_operations import TerminalOperations

class Todo:
    """Single Todo item"""

    def __init__(self, item, index,
            colored="", priority="", contexts=[], projects=[],
            creation_date="", due_date="", completed_date=""):
        self.raw            = item.strip()
        self.raw_index      = index
        self.creation_date  = creation_date
        self.priority       = priority
        self.contexts       = contexts
        self.projects       = projects
        self.due_date       = due_date
        self.completed_date = completed_date
        self.colored        = self.highlight()
        # self.colored_length = TerminalOperations.length_ignoring_escapes(self.colored)

    def update(self, item):
        self.raw            = item.strip()
        self.priority       = Todos.priority(item)
        self.contexts       = Todos.contexts(item)
        self.projects       = Todos.projects(item)
        self.creation_date  = Todos.creation_date(item)
        self.due_date       = Todos.due_date(item)
        self.completed_date = Todos.completed_date(item)
        self.colored        = self.highlight()
        # self.colored_length = TerminalOperations.length_ignoring_escapes(self.colored)

    def __repr__(self):
        return repr({
            "raw":            self.raw,
            "colored":        self.colored,
            "raw_index":      self.raw_index,
            "priority":       self.priority,
            "contexts":       self.contexts,
            "projects":       self.projects,
            "creation_date":  self.creation_date,
            "due_date":       self.due_date,
            "completed_date": self.completed_date
        })

    def highlight(self, line="", show_due_date=True, show_contexts=True, show_projects=True):
        colored = self.raw if line == "" else line
        color_list = [colored]

        if colored[:2] == "x ":
            color_list = ('completed', color_list)
        else:
            words_to_be_highlighted = self.contexts + self.projects
            if self.due_date:
                words_to_be_highlighted.append("due:"+self.due_date)
            if self.creation_date:
                words_to_be_highlighted.append(self.creation_date)

            if words_to_be_highlighted:
                color_list = re.split("(" + "|".join([re.escape(w) for w in words_to_be_highlighted]) + ")", self.raw)
                for index, w in enumerate(color_list):
                   if w in self.contexts:
                       color_list[index] = ('context', w) if show_contexts else ''
                   elif w in self.projects:
                       color_list[index] = ('project', w) if show_projects else ''
                   elif w == "due:"+self.due_date:
                       color_list[index] = ('due_date', w) if show_due_date else ''
                   elif w == self.creation_date:
                       color_list[index] = ('creation_date', w)

            if self.priority and self.priority in "ABCDEF":
                color_list = ("priority_{0}".format(self.priority.lower()), color_list)
            else:
                color_list = ("plain", color_list)

        return color_list

    def highlight_search_matches(self, line=""):
        colored = self.raw if line == "" else line
        color_list = [colored]
        if self.search_matches:
            color_list = re.split("(" + "|".join([re.escape(match) for match in self.search_matches]) + ")", self.raw)
            for index, w in enumerate(color_list):
                if w in self.search_matches:
                    color_list[index] = ('search_match', w)
        return color_list

    def is_complete(self):
        if self.raw[0:2] == "x ":
            return True
        elif self.completed_date == "":
            return False
        else:
            return True

    def complete(self):
        today = date.today()
        self.raw = "x {0} ".format(today) + self.raw
        self.completed_date = "{0}".format(today)
        self.update(self.raw)

    def incomplete(self):
        self.raw = re.sub(Todos._completed_regex, "", self.raw)
        self.completed_date = ""
        self.update(self.raw)

    def add_creation_date(self):
        if self.creation_date == "":
            p = "({0}) ".format(self.priority) if self.priority != "" else ""
            self.update("{0}{1} {2}".format(p, date.today(), self.raw.replace(p, "")))


class Todos:
    """Todo items"""
    _context_regex       = re.compile(r'(?:^|\s+)(@\S+)')
    _project_regex       = re.compile(r'(?:^|\s+)(\+\S+)')
    _creation_date_regex = re.compile(r'^'
                                      r'(?:x \d\d\d\d-\d\d-\d\d )?'
                                      r'(?:\(\w\) )?'
                                      r'(\d\d\d\d-\d\d-\d\d)\s*')
    _due_date_regex      = re.compile(r'\s*due:(\d\d\d\d-\d\d-\d\d)\s*')
    _priority_regex      = re.compile(r'\(([A-Z])\) ')
    _completed_regex     = re.compile(r'^x (\d\d\d\d-\d\d-\d\d) ')

    def __init__(self, todo_items, file_path):
        self.file_path = file_path
        self.update(todo_items)

    def reload_from_file(self):
        with open(self.file_path, "r") as todotxt_file:
            self.update(todotxt_file.readlines())

    def save(self):
        with open(self.file_path, "w") as todotxt_file:
            todotxt_file.write( "\n".join([t.raw for t in self.todo_items]) )

    def update(self, todo_items):
        self.parse_raw_entries(todo_items)

    def append(self, item, add_creation_date=True):
        self.insert(len(self.todo_items), item, add_creation_date)
        return len(self.todo_items)-1

    def insert(self, index, item, add_creation_date=True):
        self.todo_items.insert(index, self.create_todo(item, index) )
        self.update_raw_indices()
        newtodo = self.todo_items[index]
        if add_creation_date and newtodo.creation_date == "":
            newtodo.add_creation_date()
        return index

    def delete(self, index):
        del self.todo_items[index]
        self.update_raw_indices()

    def __iter__(self):
        self.index = -1
        return self

    def __next__(self):
        self.index = self.index + 1
        if self.index == len(self.todo_items):
            raise StopIteration
        return self.todo_items[self.index]

    def next(self):
        self.index = self.index + 1
        if self.index == len(self.todo_items):
            raise StopIteration
        return self.todo_items[self.index]

    def __len__(self):
        return len(self.todo_items)

    def pending_items(self):
        return [t for t in self.todo_items if not t.is_complete()]

    def done_items(self):
        return [t for t in self.todo_items if t.is_complete()]

    def pending_items_count(self):
        return len(self.pending_items())

    def done_items_count(self):
        return len(self.done_items())

    def __getitem__(self, index):
        return self.todo_items[index]

    def __repr__(self):
        return repr( [i for i in self.todo_items] )

    def create_todo(self, todo, index):
        return Todo(todo, index,
            contexts       = Todos.contexts(todo),
            projects       = Todos.projects(todo),
            priority       = Todos.priority(todo),
            creation_date  = Todos.creation_date(todo),
            due_date       = Todos.due_date(todo),
            completed_date = Todos.completed_date(todo))

    def parse_raw_entries(self, raw_items):
        self.todo_items = [
            self.create_todo(todo, index)
            for index, todo in enumerate(raw_items) if todo.strip() != ""]

    def update_raw_indices(self):
        for index, todo in enumerate(self.todo_items):
            todo.raw_index = index

    @staticmethod
    def contexts(item):
        return sorted( Todos._context_regex.findall(item) )

    @staticmethod
    def projects(item):
        return sorted( Todos._project_regex.findall(item) )

    @staticmethod
    def creation_date(item):
        match = Todos._creation_date_regex.search(item)
        return match.group(1) if match else ""

    @staticmethod
    def due_date(item):
        match = Todos._due_date_regex.search(item)
        return match.group(1) if match else ""

    @staticmethod
    def priority(item):
        match = Todos._priority_regex.match(item)
        return match.group(1) if match else ""

    @staticmethod
    def completed_date(item):
        match = Todos._completed_regex.match(item)
        return match.group(1) if match else ""

    def all_contexts(self):
        # Nested Loop
        # all_contexts = []
        # for item in self.raw_items:
        #   for found_context in self.contexts(item):
        #     if found_context not in all_contexts:
        #       all_contexts.append(found_context)
        # return all_contexts

        # List comprehension
        # return sorted(set( [found_context for item in self.raw_items for found_context in self.contexts(item)] ))

        # Join all items and use one regex.findall
        # return sorted(set( Todos._context_regex.findall(" ".join(self.raw_items))))

        return sorted(set( [context for todo in self.todo_items for context in todo.contexts] ))

    def all_projects(self):
        # List comprehension
        # return sorted(set( [project for item in self.raw_items for project in self.projects(item)] ))

        # Join all items and use one regex.findall
        # return sorted(set( Todos._project_regex.findall(" ".join(self.raw_items))))

        return sorted(set([project for todo in self.todo_items for project in todo.projects] ))

    def sorted(self, reversed_sort=False):
        self.todo_items.sort( key=lambda todo: todo.raw, reverse=reversed_sort )

    def sorted_reverse(self):
        self.sorted(reversed_sort=True)

    def sorted_raw(self):
        self.todo_items.sort( key=lambda todo: todo.raw_index )

    def swap(self, first, second):
        """
        Swap items indexed by *first* and *second*.

        Out-of-bounds situations are handled by wrapping.
        """
        if second < first:
            second, first = first, second

        n_items = len(self.todo_items)

        if first < 0:
            first += n_items

        if second >= n_items:
            second = n_items - second

        self.todo_items[first], self.todo_items[second] = self.todo_items[second], self.todo_items[first]

    def filter_context(self, context):
        return [item for item in self.todo_items if context in item.contexts]

    def filter_project(self, project):
        return [item for item in self.todo_items if project in item.projects]

    def filter_context_and_project(self, context, project):
        return [item for item in self.todo_items if project in item.projects and context in item.contexts]

    def filter_contexts_and_projects(self, contexts, projects):
        return [item for item in self.todo_items if set(projects) & set(item.projects) or set(contexts) & set(item.contexts)]

    def search(self, search_string):
        search_string = re.escape(search_string)
        # print(search_string)
        ss = []
        substrings = search_string.split("\\")
        for index, substring in enumerate(substrings):
            s = ".*?".join(substring)
            # s.replace(" .*?", " ")
            if 0 < index < len(substrings)-1:
                s += ".*?"
            ss.append(s)
        # print(repr(ss))
        search_string_regex  = '^.*('
        search_string_regex += "\\".join(ss)
        search_string_regex += ').*'
        # print(search_string_regex)

        r = re.compile(search_string_regex, re.IGNORECASE)
        results = []
        for t in self.todo_items:
            match = r.search(t.raw)
            if match:
                t.search_matches = match.groups()
                results.append(t)
        return results

    quotes = [
        "What you really believe about the source of great performance thus becomes the foundation of all you will ever achieve -- Geoff Colvin, Talent is Overrated: What Really Separates World-Class Performers from Everybody Else",
        "Do not let your good ideas to spend your brain energy, do it with all the effort because the shadow of the success can become a reality only with hard work.  -- Isra",
        "I'm exhausted. I spent all morning putting in a comma and all afternoon taking it out -- Oscar Wilde",
        "Prescribing hard work for the soft, or easy work for the hardy, is generally nonsense. What is always needed in any aim is right effort, right time, right people, right materials.  -- Idries Shah, Reflections",
        "Easy and difficult things are juat small parts of life, the rest are what have to be done.  -- Alief Moulana",
        "Hard work is much more important than talent -- Carlo Rotella",
        "Wrote my way out of the hood...thought my way out of poverty! Don't tell me that knowledge isn't power. Education changes everything.  -- Brandi L. Bates",
        "Everything yields to diligence -- Thomas Jefferson",
        "Embrace the pain to inherit the gain.  -- Habeeb Akande",
        "If you don't burn out at the end of each day, you're a bum.  -- George Lois",
        "If you work hard and study hard. And you fuck up. That's okay. If you fuck up and you fuck up, then you're a fuckup -- Justin Halpern, Sh*t My Dad Says",
        "Patience can be bitter but her fruit is always sweet.  -- Habeeb Akande",
        "Greatness is sifted through the grind, therefore don't despise the hard work now for surely it will be worth it in the end.  -- Sanjo Jendayi",
        "You know how when you step on court your coach is like 'go go go!'? And all throughout you just keep telling yourself to hit harder and harder and keep at it? You know how much you treasure those five-minute timeouts? You know how good you feel at the end of a session? You know how you're glad you're tired? No pills, no shots, just plain energy. I want to work like that. Whether I have to write ten thousand words or send five hundred emails, brainstorm for hours at a time, I want to have that energy. To keep fighting. To know it's all worth it.  -- Thisuri Wanniarachchi",
        "Many say'What goes up must come down.' I believe this was spoken by a pessimist. I believe it is vitally important to remember that'What goes up...must have started some place.d -- Jayce O'Neal",
        "When you are pursuing your dreams, they will call you CRAZY because they are LAZY. They never know you are a HERO who just jumped away from step ZERO.  Stay away from negative people; they will only pollute you.  -- Israelmore Ayivor",
        "learning texts is worth doing not because it's easy but because it's hard.  -- Joshua Foer, Moonwalking with Einstein: The Art and Science of Remembering Everything",
        "Hard-earned good money is hard to waste; but can still have some good moments.  -- POB Bismark",
        "Talent is what God gives us, Skill is what we give back to Him -- Eliel Pierre",
        "Work as if there's no one to help you, and learn as if everybody is with you.  -- saransh garg",
        "To achieve what 1% of the worlds population has (Financial Freedom), you must be willing to do what only 1% dare to do..hard work and perseverance of highest order.  -- Manoj Arora, From the Rat Race to Financial Freedom",
        "Believe that your hard work, dedication and persistence will pay off; improve through continual learning and believe in your future.  -- Lorii Myers, No Excuses, The Fit Mind-Fit Body Strategy Book",
        "There is an open circle. -- John Passaro",
        "Love what you do, do it with passion, work hard, and never give up. Money and happiness will come to you.  -- Joseph C. Kunz Jr. ",
        "Only cooked time tastes well -- Zeeshan Ahmed",
        "Sometimes the greatest thing to come out of all your hard work isn't what you GET for it, but what you BECOME for it.  -- Steve Maraboli, Unapologetically You: Reflections on Life and the Human Experience",
        "Be Patient to become a Patent -- Zeeshan Ahmed",
        "I dropped out of school and congratulated myself for my diligence. Few realize how hard one has to work to resist the pressures of conventional success.  -- Bauvard, The Prince Of Plungers",
        "I need to dream.  -- John Passaro, 6 Minutes Wrestling With Life",
        "I'm lazy! I hate work! Hate hard work in all its forms! Clever shortcuts, that's all I'm about!  -- Eliezer Yudkowsky, Harry Potter and the Methods of Rationality",
        "We must strive for literacy and education that teach us to never quit questioning and probing at the assumptions of the day.  -- Bryant McGill, Voice of Reason",
        "I have learned that real angels don't have gossamer white robes and Cherubic skin, they have calloused hands and smell of the days' sweat.  -- Richard Paul Evans, Lost December",
        "Success necessitates sacrifice.  -- Habeeb Akande",
        "You always need to work hard.  You always need to be willing to work hard.  Not everything will be hard, but you should, at the very least, be willing to work hard.  -- Tom Giaquinto, Be A Good Human",
        "Do the things you like to be happier, stronger & more successful. Only so is hard work replaced by dedication.  -- Rossana Condoleo",
        "There comes a point when you have to realize that the sum of all your blood, sweat, and tears will ultimately amount to zero.  -- Max Brooks, World War Z: An Oral History of the Zombie War",
        "There is something beautiful in you seeking freedom.  -- Bryant McGill, Voice of Reason",
        "Be like a duck, paddling and working very hard inside the water, but what everyone sees is a smiling and calm face.  -- Manoj Arora, From the Rat Race to Financial Freedom",
        "If you are not working towards something, your life will end with nothing.  -- Habeeb Akande",
        "Respect your dream.  -- Jill Williamson, Go Teen Writers: How to Turn Your First Draft into a Published Book",
        "Hard work beats talent when talent fails to work hard.  -- Kevin Durant",
        "Im not a very good sleeper. But you know what? I'm willing to put in a few extra hours every day to get better. That's just the kind of hard worker I am.  -- Jarod Kintz, Whenever You're Gone, I'm Here For You",
        "Persistence and passion will make you invincible.  -- Christian Baloga",
        "When God calls you to build 100 castles on earth and you built 98, take the 99th as if it's the begining of your work and work hard to finish the race with all excellence. Go the extra mile!  -- Israelmore Ayivor",
        # "It's a complex song, and it's fascinating to watch the creative process as they went back and forth and finally created it over a few months. Lennon was always my favorite Beatle. [ He laughs as Lennon stops during the first take and makes the band go back and revise a chord.] Did you hear that little detour they took? It didn't work, so they went back and started from where they were. It's so raw in this version. It actually makes the sound like mere mortals. You could actually imagine other people doing this, up to this version. Maybe not writing and conceiving it, but certainly playing it. Yet they just didn't stop. They were such perfectionists they kept it going This made a big impression on me when I was in my thirties. You could just tell how much they worked at this.  -- Walter Isaacson, Steve Jobs",
        "All success comes down to this . . . action -- Rob Liano",
        "It is very rare for a child of God to find gold and crude oil on the floor to fetch. He/she must dig and dig deeply well!!!  -- Israelmore Ayivor",
        "When the grass is greener at other people's feet, it is not because the grass chose to take up that complexion. But it is because, they have deliberately irrigated it on regular accounts.  -- Israelmore Ayivor",
        "Some people dream of SUCCESS while other wake up and work hard at it.  -- Barbara Rubel",
        "If you want a fried fish to fly and enter your mouth, you must keep waiting till the unending time ends. Dead fish doesn't fly. If you want to eat it, your own hands must carry it.  -- Israelmore Ayivor",
        "Loosers want it but only through the easy way. If action taking were like reading a story book, loosers would only love to open the picture pages -- Israelmore Ayivor",
        "Virginity comes standard. A good head is earned.  -- Mokokoma Mokhonoana",
        "I've never felt that the American Dream was owed to me. I never felt that I was entitled to this Dream. This is why I laid the cobblestone before me for this Dream to be achieved.  -- John-Talmage Mathis, I Deal to Plunder - A ride through the boom town",
        "As I tell my children, 'If you are going to do something, do your best while you're doing it.  -- Michelle    Moore, Selling Simplified",
        "The more civilized people are, the more honorable working hard is to them. As a result, the more civilized we get, the less we live.  -- Mokokoma Mokhonoana",
        "Children take joy in their work and sometimes as adults we forget that's something we should continue doing.  -- Ashley Ormon, God in Your Morning",
        "If you want your dreams to work out for you, you must work with them. Pay the price and have the package of your accomplishments in full versions.  -- Israelmore Ayivor",
        "Nonetheless, the fact remains; he had hope in a better world he could not yet see that overwhelmed the cries of'you can't' or'you won't' or'why bother.'   More than anything else, mastering that faith, on cue, is what separated him from his peers, and distinguishes him from so many people in these literal, sophisticated times.   It has made all the difference.  -- Ron Suskind, A Hope in the Unseen: An American Odyssey from the Inner City to the Ivy League",
        "Ideas do not work..  -- Manoj Arora, From the Rat Race to Financial Freedom",
        "Stop and say something good about yourself! Believe what you have just said! Pray for what you have just believed! Have faith for what you just prayed for! Work out that faith! ...and surely goodness and mercy shall follow you!  -- Israelmore Ayivor",
        "It may be that you will be happiest in the rat race; perhaps, like me, you are basically a rat.  -- Richard Koch, The 80/20 Principle: The Secret to Achieving More with Less",
        "The world is moved along, not only by the mighty shoves of its heroes, but also by the aggregate of tiny pushes of each honest worker.  -- Helen Keller",
        "Your dreams are like the cement. If you water it with actions, it becomes a hard concrete mass. But if you leave it exposed and unwatered, the air will easily blow it away!  -- Israelmore Ayivor",
        "You may have the greatest vision, plans or goals as you may term it. You can call it Vison 2020, Vision 2045 or whatever. But remember, not work is done unless a distance is covered!  -- Israelmore Ayivor",
        "As I see it, if you work more hours than somebody else, during those hours you learn more about your craft. That can make you more efficient, more able, even happier. Hard work is like compounded interest in the bank. The rewards build faster.  -- Randy Pausch, The Last Lecture",
        "For a great tomorrow, plan today. Because every tomorrow becomes today, and every today becomes tomorrow. -- Richmond Akhigbe",
        "This should have been a red flag, I realize in retrospect. Working really hard on anything is, by definition, not cool.  -- Leila Sales, This Song Will Save Your Life",
        "I want the peace in knowing that is wasn't for lack of hustling that I missed a target for my dream. I want to know that the one thing in my control was under control.  -- Jon Acuff, Quitter: Closing the Gap Between Your Day Job and Your Dream Job",
        "Though you can love what you do not master, you cannot master what you do not love.  -- Mokokoma Mokhonoana",
        "Today's marriages become toxic, with resentments, after only a few years. It's one thing to say, 'I forgive,' but most lack the enterprise to do the necessary work that follows. It was the day after that proved who had the wisdom of God and who didn't.  -- Michael Ben Zehabe, Song of Songs the book for daughters",
        "I believe in success.  -- Himmilicious",
        "The pretty ones are usually unhappy. They expect everyone to be enamored of their beauty. How can a person be content when their happiness lies in someone else's hands, ready to be crushed at any moment? Ordinary-looking people are far superior, because they are forced to actually work hard to achieve their goals, instead of expecting people to fall all over themselves to help them.  -- J. Cornell Michel, Jordan's Brains: A Zombie Evolution",
        "I can only strive for what is important -- Rosie Thomas, Iris And Ruby",
        "To be a great achiever, you must first discover your main purpose of existing through feasible communication mode with your maker, the source of your destiny. This brings you a conviction to work with.  -- Israelmore Ayivor, Michelangelo | Beethoven | Shakespeare: 15 Things Common to Great Achievers",
        "A block won't move without the effort.  -- Isabella Poretsis",
        "People have struggled for the benefits of others, you can struggle at least for your own benefit.  -- Amit Kalantri",
        "Man was designed in a way in which he must eat in order to give him a solid reason to go to work everyday. This helps to keep him out of trouble. God is wise.  -- Criss Jami",
        "Submitting seemed to me a lot like giving up. If God gave us the strength to bail- the gumption to try and save ourselves- isn't that what he wanted us to do?  -- Jeannette Walls, Half Broke Horses",
        "Too many irons, not enough fire.  -- S. Kelley Harrell",
        "Ah? A small aversion to menial labor?' The doctor cocked an eyebrow.'Understandable, but misplaced. One should treasure those hum-drum tasks that keep the body occupied but leave the mind and heart unfettered.  -- Tad Williams, The Dragonbone Chair",
        "I'm really very self-confident when it comes to my work. When I take on a project, I believe in it 100%. I really put my soul into it. I'd die for it. That's how I am -- Michael  Jackson",
        "Create with the heart; build with the mind.  -- Criss Jami",
        "Praying without working is faith inaction.  -- Saji Ijiyemi",
        "Such is life. It is no cleaner than a kitchen; it reeks like a kitchen; and if you mean to cook your dinner, you must expect to soil your hands; the real art is in getting them clean again, and therein lies the whole morality of our epoch.  -- Honore de Balzac, Pere Goriot",
        "It's a question of attitude. If you really work at something you can do it up to a point. If you really work at being happy you can do it up to a point. But anything more than that you can't. Anything more than that is luck.  -- Haruki Murakami, Dance Dance Dance",
        "With intellectual labor your hard work is forever, while with manual labor your hard work is temporary and soon forgotten.  -- Jarod Kintz, At even one penny, this book would be overpriced. In fact, free is too expensive, because you'd still waste time by reading it. ",
        "I feel as though whenever I create something, my Mr. Hyde wakes up in the middle of the night and starts thrashing it. I sometimes love it the next morning, but other times it is an abomination.  -- Criss Jami",
        "Does the work get easier once you know what you are doing?' -- Christopher Moore, Lamb: The Gospel According to Biff, Christ's Childhood Pal",
        "All good work requires self-revelation.  -- Sidney Lumet, Making Movies",
        "I've always resented the smug statements of politicians, media commentators, corporate executives who talked of how, in America, if you worked hard you would become rich.  The meaning of that was if you were poor it was because you hadn't worked hard enough.  I knew this was a lite, about my father and millions of others, men and women who worked harder than anyone, harder than financiers and politicians, harder than anybody if you accept that when you work at an unpleasant job that makes it very hard work indeed.  -- Howard Zinn, You Can't Be Neutral on a Moving Train: A Personal History of Our Times",
        "Hard work without talent is a shame, but talent without hard work is a tragedy -- Robert Half",
        "Sometimes it takes a lowly, title-less man to humble the world. Kings, rulers, CEOs, judges, doctors, pastors, they are already expected to be greater and wiser.  -- Criss Jami, Venus in Arms",
        "If your dream is a big dream, and if you want your life to work on the high level that you say you do, there's no way around doing the work it takes to get you there.  -- Joyce Chapman",
        "I do not care about happiness simply because I believe that joy is something worth fighting for.  -- Criss Jami",
        "The writer's curse is that even in solitude, no matter its duration, he never grows lonely or bored.  -- Criss Jami",
        "If wealth was the inevitable result of hard work and enterprise, every woman in Africa would be a millionaire.  -- George Monbiot",
        "I'm really very self-confident when it comes to my work. When I take on a project, I believe in it 100%. I really put my soul into it. I'd die for it. That's how I am.  -- Michael  Jackson",
        "In the land where excellence is commended, not envied, where weakness is aided, not mocked, there is no question as to how its inhabitants are all superhuman.  -- Criss Jami, Venus in Arms",
        "No one ever drowned in sweat.  -- United States Marine Corps",
        "What is hard work? It takes strength, energy, and stress to truly care about others enough to place oneself last, but it is easy to wrap oneself up and selfishly scramble on the heads of others.  -- Criss Jami",
        "I don't have a blue-collar job. It's more of a green collar, because of all the yellow sweat stains mixing in.   -- Jarod Kintz, This Book Has No Title",
        "As Aristotle said, 'Excellence is a habit.' I would say furthermore that excellence is made constant through the feeling that comes right after one has completed a work which he himself finds undeniably awe-inspiring. He only wants to relax until he's ready to renew such a feeling all over again because to him, all else has become absolutely trivial.  -- Criss Jami",
        "Inspiration is the windfall from hard work and focus. Muses are too unreliable to keep on the payroll.  -- Helen Hanson",
        "The idea that the harder you work, the better you're going to be is just garbage. The greatest improvement is made by the man or woman who works most intelligently.  -- Bill Bowerman",
        "You were hired because you met expectations, you will be promoted if you can exceed them.  -- Saji Ijiyemi",
        "People.. were poor not because they were stupid or lazy. They worked all day long, doing complex physical tasks. They were poor because the financial institution in the country did not help them widen their economic base.  -- Muhammad Yunus, Banker to the Poor: Micro-Lending and the Battle Against World Poverty",
        "She was tough in the best sense of the word.  She'd taken blows, the disappointments, and had worked her way through them.  Some people, he knew, would have buckled under, found a clutch, or given up.  But she had carved a place for herself and made it work.  -- Nora Roberts",
        "A clay pot sitting in the sun will always be a clay pot.  It has to go through the white heat of the furnace to become porcelain.  -- Mildred W. Struven",
        "Barking hard work, being boy.  -- Scott Westerfeld, Leviathan",
        "Do I attribute my success to hard work, or sunscreen? If you want the truth, maybe you should ask my new albino secretary.   -- Jarod Kintz, At even one penny, this book would be overpriced. In fact, free is too expensive, because you'd still waste time by reading it. ",
        "Sometimes there's not a better way. Sometimes there's only the hard way.  -- Mary E. Pearson, The Fox Inheritance",
        "The harder you fall, the heavier your heart; the heavier your heart, the stronger you climb; the stronger you climb, the higher your pedestal.  -- Criss Jami",
        "Every job from the heart is, ultimately, of equal value. The nurse injects the syringe; the writer slides the pen; the farmer plows the dirt; the comedian draws the laughter. Monetary income is the perfect deceiver of a man's true worth.  -- Criss Jami",
        "It is a pity that doing one's best does not always answer.  -- Charlotte Brontë, Jane Eyre",
        "I don't like to celebrate my birthday, because I don't like taking credit for others' work—in this case, my mom and dad. Or possibly my mom and the mailman.  -- Jarod Kintz, This Book Has No Title",
        "The dictionary is the only place that success comes  before work. work is the key to success, and hard work can help you accomplish anything.  -- Vince Lombardi",
        "Many who are self-taught far excel the doctors, masters, and bachelors of the most renowned universities.  -- Ludwig von Mises",
        "...talent means nothing, while experience, acquired in humility and with hard work, means everything.  -- Patrick Süskind, Perfume: The Story of a Murderer",
        "No one understands and appreciates the American Dream of hard work leading to material rewards better than a non-American.  -- Anthony Bourdain, Kitchen Confidential: Adventures in the Culinary Underbelly",
        "The difference between ordinary and extraordinary is that little extra.  -- Jimmy Johnson",
        "The year you were born marks only your entry into the world. Other years where you prove your worth, they are the ones worth celebrating.   -- Jarod Kintz, This Book Title is Invisible",
        "Don't wish it were easier. Wish you were better.  -- Jim Rohn",
        "There are no shortcuts to any place worth going.  -- Beverly Sills",
        "Determine never to be idle.  No person will have occasion to complain of the want of time, who never loses any.  It is wonderful how much may be done, if we are always doing.  -- Thomas Jefferson",
        "The three great essentials to achieve anything worthwhile are, first, hard work; second, stick-to-itiveness; third, common sense.  -- Thomas A. Edison",
        "I'm a greater believer in luck, and I find the harder I work the more I have of it -- Thomas Jefferson",
        "If you try and lose then it isn't your fault. But if you don't try and we lose, then it's all your fault.  -- Orson Scott Card, Ender's Game",
        "Make a pact with yourself today to not be defined by your past. Sometimes the greatest thing to come out of all your hard work isn't what you get for it, but what you become for it. Shake things up today! Be You...Be Free...Share.  -- Steve Maraboli, Life, the Truth, and Being Free",
        "If I told you I've worked hard to get where I'm at, I'd be lying, because I have no idea where I am right now.  -- Jarod Kintz, This Book is Not for Sale",
        "It's hard to beat a person who never gives up.  -- Babe Ruth",
        "The best way to not feel hopeless is to get up and do something. Don't wait for good things to happen to you. If you go out and make some good things happen, you will fill the world with hope, you will fill yourself with hope. -- Barack Obama",
        "The only man who never makes mistakes is the man who never does anything. --  Theodore Roosevelt",
        "What would life be if we had no courage to attempt anything? -- Vincent van Gogh",
        "Far better it is to dare mighty things, to win glorious triumphs, even though checkered by failure, than to take rank with those poor spirits who neither enjoy much nor suffer much, because they live in the gray twilight that knows neither victory nor defeat. -- Theodore Roosevelt, Strenuous Life",
        "...wanting change is step one, but step two is taking it. -- Isaac Marion, Warm Bodies",
        "Make voyages. Attempt them. There's nothing else. -- Tennessee Williams, Camino Real",
        "Do more than belong: participate. Do more than care: help. Do more than believe: practice. Do more than be fair: be kind. Do more than forgive: forget. Do more than dream: work. -- William Arthur Ward",
        "To live a creative life, we must lose our fear of being wrong. -- Joseph Chilton Pearce",
        "I love to sail forbidden seas, and land on barbarous coasts. -- Herman Melville",
        "Do anything, save to lie down and die! -- Nathaniel Hawthorne, The Scarlet Letter",
        "Whatever the mind of man can conceive and believe, it can achieve -- W. Clement Stone",
        "He who waits to do a great deal of good at once will never do anything. -- Samuel Johnson",
        "There's doubt in trying.  Just do it or stop thinking. -- Toba Beta, Master of Stupidity",
        "You'll never know the the outcome unless you just go out there and just do it. -- Jonathan Anthony Burkett, Neglected But Undefeated: The Life Of A Boy Who Never Knew A Mother's Love",
        "The only dreams that matter are the ones you have when you're awake. -- John O'Callaghan",
        "A person can choose to stand against adversity, but most people will run from it before they come to terms with who they truly are, a wise person watches through adversity and follows the signs when it is their turn to stand. -- Faith Brashear",
        "People tell you that you cannot, because they do not. -- Tim Fargo",
    ]

    @staticmethod
    def quote():
        return Todos.quotes[ random.randrange(len(Todos.quotes)) ]



#!/usr/bin/env python
# coding=utf-8
import sys
# import os
import subprocess
import tty
import termios
import re
# import curses
import fcntl
import struct

class TerminalOperations:
    """For interacting with the terminal"""

    _escape_sequence_regex = re.compile(r'\x1b\[[0-9;]*m')
    _screen_size_regex     = re.compile(r'\[8;(.*);(.*)t')

    @staticmethod
    def foreground_color(index):
        return "\x1B[38;5;{0}m".format(index)

    @staticmethod
    def background_color(index):
        return "\x1B[48;5;{0}m".format(index)

    @staticmethod
    def clear_formatting():
        return "\x1B[m"

    def __init__(self, use_tput=False):
        # self.window = curses.initscr()
        self.update_screen_size(use_tput)

    def update_screen_size(self, use_tput=False):
        self.columns, self.rows = self.screen_size(use_tput)

    def output(self, text):
        sys.stdout.write(text)

    def hide_cursor(self):
        # subprocess.check_output(["tput", "civis"])
        self.output('\x1b[?25l')

    def show_cursor(self):
        # subprocess.check_output(["tput", "cnorm"])
        self.output('\x1b[34h\x1b[?25h')

    def clear_screen(self):
        self.output("\x1B[2J")

    def screen_size(self, use_tput=False):
        # Method: Usint tput
        if use_tput:
            return ( int(subprocess.check_output(["tput", "cols"])), int(subprocess.check_output(["tput", "lines"])) )
        else:
            # this is how urwid does it in raw_terminal display mode
            buf = fcntl.ioctl(0, termios.TIOCGWINSZ, ' '*4)
            y, x = struct.unpack('hh', buf)
            return x, y

        # Query the terminal directly in bash:
        # stty -echo; echo -en "\033[18t"; read -d t size; stty echo; size=${size/#??/}; echo $size

        # Method: querying the terminal directly with escape sequences
        # the following breaks if the terminal does not have this capability, ie if TERM=screen*
        # response = ""
        # c        = ""
        # if set_terminal_raw:
        #     original_stdin_settings = termios.tcgetattr(sys.stdin.fileno())
        #     tty.setraw(sys.stdin.fileno())
        # sys.stdout.write("\x1B[18t")
        # try:
        #     while c != "t":
        #         c = sys.stdin.read(1)
        #         response += c
        # finally:
        #     if set_terminal_raw:
        #         termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, original_stdin_settings)
        # response = response.split(";")
        # rows     = int(response[-2])
        # columns  = int(response[-1][0:-1])
        # return (columns, rows)

        # Method using curses module
        # return self.window.getmaxyx()


    def move_cursor(self, row, column):
        self.output("\x1B[{0};{1}H".format(row, column))

    def move_cursor_home(self):
        self.output("\x1B[[H")

    def move_cursor_next_line(self):
        self.output("\x1B[E")

    @staticmethod
    def length_ignoring_escapes(line):
        return len(line) - sum([len(i) for i in TerminalOperations._escape_sequence_regex.findall(line)])

    @staticmethod
    def ljust_with_escapes(line, columns, string_length=0):
        length = TerminalOperations.length_ignoring_escapes(line) if string_length == 0 else string_length
        if length < columns:
            line += " " * (columns - length)
        return line

    # solution for single key press - blocking
    # def getch(self):
    #     """getch() -> key character

    #     Read a single keypress from stdin and return the resulting character.
    #     Nothing is echoed to the console. This call will block if a keypress
    #     is not already available, but will not wait for Enter to be pressed.

    #     If the pressed key was a modifier key, nothing will be detected; if
    #     it were a special function key, it may return the first character of
    #     of an escape sequence, leaving additional characters in the buffer.
    #     """
    #     fd = sys.stdin.fileno()
    #     old_settings = termios.tcgetattr(fd)
    #     try:
    #         tty.setraw(fd)
    #         ch = sys.stdin.read(1)
    #     finally:
    #         termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    #     return ch

# print("\x1B]0;THIS IS A TITLE BAR DEMO...\x07")
# print("Wait for 5 seconds...")
# print("\x1B]0;\x07")

# def screen_size_using_escapes
#   state = `stty -g`
#   `stty raw -echo -icanon isig`
#   STDOUT.write "\e[18t"
#   response = c = ""
#   while c != 't'
#     c = STDIN.getbyte.chr #if STDIN.ready?
#     response += c.to_s unless c == "\e"
#   end
#   if response =~ /\[8;(.*);(.*)t/
#     rows = $1
#     cols = $2
#     return [cols, rows]
#   end
#   return []
# ensure
#   `stty #{state}`
# end

# Ruby - get single keypress
# def get_char
#   state = `stty -g`
#   `stty raw -echo -icanon isig`
#   STDIN.getc.chr
# ensure
#   `stty #{state}`
# end




#!/usr/bin/env python
# coding=utf-8

import os
import sys

# Import the correct version of configparser
if sys.version_info[0] >= 3:
    import configparser
    config_parser_module = configparser
elif sys.version_info[0] < 3:
    import ConfigParser
    config_parser_module = ConfigParser

class ColorScheme:

    def __init__(self, name, user_config):
        self.built_in_colors_directory = os.path.realpath(os.path.expanduser(os.path.dirname(__file__) + '/colors'))
        self.user_config = user_config
        self.load_colors(name)

    def load_colors(self, name):
        self.colors = {}
        self.focus_map = {}
        self.dialog_focus_map = {}
        colorscheme_section = "colorscheme-{0}".format(name)

        # Use user defined theme in the user_config if it exists
        if self.user_config.has_section(colorscheme_section):
            self.colors = dict( self.user_config.items(colorscheme_section) )
        else:
            # Try to load a built in theme
            cfg = config_parser_module.ConfigParser()
            if name in os.listdir(self.built_in_colors_directory):
                cfg.read(self.built_in_colors_directory + "/" + name)
            # Load default theme
            else:
                cfg.read(self.built_in_colors_directory + "/default")
                colorscheme_section = "colorscheme-default"
            if cfg.has_section(colorscheme_section):
                self.colors = dict( cfg.items(colorscheme_section) )

        # Split foreground and background values
        for key, value in self.colors.items():
            color_strings = value.split(',')
            if len(color_strings) == 1:
                color_strings.append('')
            self.colors[key] = {'fg': color_strings[0], 'bg': color_strings[1]}

        # Create Selected attributes using the selected_background_color
        selected_background_color = self.colors['selected']['bg']
        dialog_color = self.colors['dialog_color']['bg']
        # dialog_button_color = self.colors['dialog_button_color']['bg']
        for key, value in list(self.colors.items()):
            if key not in ['selected', 'dialog_color', 'dialog_button_color']:
                self.colors[key+'_selected'] = {'fg': self.colors[key]['fg'], 'bg': selected_background_color}
                self.colors[key+'_dialog_color'] = {'fg': self.colors[key]['fg'], 'bg': dialog_color}
                # self.colors[key+'_dialog_button_color'] = {'fg': self.colors[key]['fg'], 'bg': dialog_button_color}
                self.focus_map[key] = key + '_selected'
                self.dialog_focus_map[key] = key + '_dialog_color'
                # self.dialog_focus_map[key] = key + '_dialog_button_color'



#!/usr/bin/env python
# coding=utf-8
import pytest
from .. import terminal_operations

@pytest.fixture
def term():
  return terminal_operations.TerminalOperations(use_tput=True)

def test_terminal_operations_init(term):
  assert type(term.rows) == int
  assert type(term.columns) == int
  assert term.rows > 0
  assert term.columns > 0

def test_terminal_operations_foreground_color(term):
  for i in range(0, 256):
    assert term.foreground_color(i) == "\x1B[38;5;{}m".format(i)

def test_terminal_operations_move_cursor(term, capsys):
  for i in range(0, 5):
    term.move_cursor(i, i)
    term.output("X")
  out, error = capsys.readouterr()
  assert out == "\x1B[0;0HX\x1B[1;1HX\x1B[2;2HX\x1B[3;3HX\x1B[4;4HX"

def test_terminal_operations_length_ignoring_escapes(term):
    assert len("(A) 2013-10-25 This is a +Very @cool test") == term.length_ignoring_escapes("\x1b[m(A) \x1b[38;5;2m2013-10-25\x1b[38;5;13m This is a \x1b[38;5;4m+Very\x1b[38;5;13m \x1b[38;5;1m@cool\x1b[38;5;13m test")

def test_terminal_operations_ljust_with_escapes(term):
    test_string = "(A) 2013-10-25 This is a +Very @cool test"
    test_string_with_escapes = "\x1b[m(A) \x1b[38;5;2m2013-10-25\x1b[38;5;13m This is a \x1b[38;5;4m+Very\x1b[38;5;13m \x1b[38;5;1m@cool\x1b[38;5;13m test"
    assert len(test_string.ljust(80)) - len(test_string) == len(term.ljust_with_escapes(test_string_with_escapes, 80)) - len(test_string_with_escapes)



# -*- coding: utf-8 -*-

"""
todotxt-machine
===============

An interactive terminal based todo.txt file editor with an
interface similar to [mutt](http://www.mutt.org/).

Requirements
------------

Python 2.7 or Python 3.3 with readline support.

Documentation
-------------

The documentation for ``todotxt-machine`` is `available on github <https://github.com/AnthonyDiGirolamo/todotxt-machine>`_.

License
-------

``todotxt-machine`` is licensed under a GPLv3 license, see ``LICENSE`` for details.
"""

from setuptools import setup, find_packages
from sys import version_info

from setuptools.command.test import test as TestCommand
import sys

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

import todotxt_machine

NAME = "todotxt-machine"

try:
    long_description = open("README.rst").read()
except IOError:
    long_description = __doc__


setup(name=NAME,
      version=todotxt_machine.version,
      author="Anthony DiGirolamo",
      author_email="anthony.digirolamo@gmail.com",
      url="https://github.com/AnthonyDiGirolamo/todotxt-machine",
      description="An interactive terminal based todo.txt file editor with an interface similar to mutt",
      long_description=long_description,
      keywords="todotxt, todo.txt, todo, terminal, urwid, curses, console",
      packages=find_packages(exclude=["todotxt_machine/test*"]),
      include_package_data=True,
      entry_points={
          'console_scripts':
            ['todotxt-machine = todotxt_machine.cli:main']
      },
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Console :: Curses",
          "Intended Audience :: End Users/Desktop",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Natural Language :: English",
          "Operating System :: POSIX :: Linux",
          "Operating System :: MacOS",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.3",
          "Topic :: Office/Business :: Scheduling",
      ],
      install_requires=['setuptools', 'docopt>=0.6.2', 'urwid>=1.2.1'],
      tests_require=['pytest'],
      cmdclass = {'test': PyTest},
      )



#!/usr/bin/env python
# coding=utf-8
import pytest
from datetime import date
from .. import todo

import pprint
pp = pprint.PrettyPrinter(indent=4).pprint

@pytest.fixture
def todos():
    return todo.Todos([
        "(A) Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ], './todo.txt')

@pytest.fixture
def today():
    return date.today()

def test_todos_init(todos):
    assert len(todos) == 5
    assert len(todos.todo_items) == 5

def test_todos_parse_entries(todos):
    todo = todos.todo_items[0]
    assert todo.raw      == "(A) Thank Mom for the dinner @phone"
    assert todo.contexts == ["@phone"]
    assert todo.projects == []
    assert todo.priority == "A"

    todo = todos.todo_items[1]
    assert todo.raw      == "(B) Schedule Goodwill pickup +GarageSale @phone"
    assert todo.contexts == ["@phone"]
    assert todo.projects == ["+GarageSale"]
    assert todo.priority == "B"

    todo = todos.todo_items[2]
    assert todo.raw      == "Unpack the guest bedroom +Unpacking due:2013-10-20"
    assert todo.contexts == []
    assert todo.projects == ["+Unpacking"]
    assert todo.due_date == "2013-10-20"

    todo = todos.todo_items[3]
    assert todo.raw           == "2013-10-19 Post signs around the neighborhood +GarageSale"
    assert todo.contexts      == []
    assert todo.projects      == ["+GarageSale"]
    assert todo.creation_date == "2013-10-19"

    todo = todos.todo_items[4]
    assert todo.raw            == "x 2013-10-01 @GroceryStore Eskimo pies"
    assert todo.contexts       == ["@GroceryStore"]
    assert todo.projects       == []
    assert todo.completed_date == "2013-10-01"

def test_todos_iterable(todos):
    for todo in todos:
        assert todo.raw != ""
    for todo in todos:
        assert todo.raw != ""

def test_todos_contexts(todos):
    assert "@phone" in todos.contexts("(A) Thank Mom for the meatballs @phone")
    assert ["@home", "@phone"] == todos.contexts("Make phonecalls from home @phone @home")

def test_todos_projects(todos):
    assert "+GarageSale" in todos.projects("(B) Schedule Goodwill pickup +GarageSale @phone")
    assert ["+deck", "+portch"] == todos.projects("Finish outdoor projects +portch +deck")

def test_context_project_regex(todos):
    todos.update([
        "(A) 1999-12-24 Thank Mom for the dinner @phone @email mom@email.com",
        "(B) Schedule Goodwill pickup +GarageSale NotA+Project @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])
    todo = todos[0]
    assert todo.raw      == "(A) 1999-12-24 Thank Mom for the dinner @phone @email mom@email.com"
    assert todo.contexts == ["@email", "@phone"]
    assert todo.projects == []
    assert todo.priority == "A"
    todo = todos[1]
    assert todo.raw      == "(B) Schedule Goodwill pickup +GarageSale NotA+Project @phone"
    assert todo.contexts == ["@phone"]
    assert todo.projects == ["+GarageSale"]
    assert todo.priority == "B"

def test_todos_all_contexts(todos):
    assert ["@GroceryStore", "@phone"] == todos.all_contexts()

def test_todos_all_projects(todos):
    assert ["+GarageSale", "+Unpacking"] == todos.all_projects()

def test_todos_completed_date(todos):
    assert todos.completed_date("2011-03-02 Document +TodoTxt task format")                  == ""
    assert todos.completed_date("(A) 2011-03-02 Document +TodoTxt task format")              == ""
    assert todos.completed_date("x 2012-03-03 2011-03-02 Document +TodoTxt task format")     == "2012-03-03"
    assert todos.completed_date("x 2012-03-03 (A) 2011-03-02 Document +TodoTxt task format") == "2012-03-03"

def test_todos_creation_date(todos):
    assert todos.creation_date("2011-03-02 Document +TodoTxt task format")                  == "2011-03-02"
    assert todos.creation_date("(A) 2011-03-02 Document +TodoTxt task format")              == "2011-03-02"
    assert todos.creation_date("x 2012-03-03 2011-03-02 Document +TodoTxt task format")     == "2011-03-02"
    assert todos.creation_date("x 2012-03-03 (A) 2011-03-02 Document +TodoTxt task format") == "2011-03-02"

def test_todos_due_date(todos):
    assert todos.due_date("2011-03-02 Document +TodoTxt task format due:2013-10-25") == "2013-10-25"
    assert todos.due_date("2011-03-02 due:2013-10-25 Document +TodoTxt task format") == "2013-10-25"
    # with pytest.raises(todo.NoDueDateError):
    assert todos.due_date("2011-03-02 Document +TodoTxt task format") == ""

def test_todos_priority(todos):
    assert todos.priority("(A) Priority A") == "A"
    assert todos.priority("(Z) Priority Z") == "Z"
    assert todos.priority("(a) No Priority") == ""
    # with pytest.raises(todo.NoPriorityError):
    assert todos.priority("No Priority (A)") == ""
    assert todos.priority("(A)No Priority") == ""
    assert todos.priority("(A)->No Priority") == ""

def test_todos_sorted(todos):
    todos.parse_raw_entries([
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "(A) Thank Mom for the dinner @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])
    assert [todo.raw_index for todo in todos.todo_items] == [0, 1, 2, 3, 4]

    todos.sorted()
    assert [todo.raw for todo in todos.todo_items] == [
        "(A) Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "x 2013-10-01 @GroceryStore Eskimo pies" ]
    assert [todo.raw_index for todo in todos.todo_items] == [1, 0, 3, 2, 4]

    todos.sorted_raw()
    assert [todo.raw for todo in todos.todo_items] == [
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "(A) Thank Mom for the dinner @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ]
    assert [todo.raw_index for todo in todos.todo_items] == [0, 1, 2, 3, 4]

def test_todos_sorted_reverese(todos):
    todos.sorted_reverse()
    assert [todo.raw for todo in todos.todo_items] == [
        "x 2013-10-01 @GroceryStore Eskimo pies",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "(A) Thank Mom for the dinner @phone" ]
    assert [todo.raw_index for todo in todos.todo_items] == [4, 2, 3, 1, 0]

def test_todos_filter_context(todos):
    assert [t.raw for t in todos.filter_context("@phone")] == [
        "(A) Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone"]
    assert [t.raw for t in todos.filter_context("@GroceryStore")] == [
        "x 2013-10-01 @GroceryStore Eskimo pies" ]

def test_todos_filter_project(todos):
    assert [t.raw for t in todos.filter_project("+GarageSale")] == [
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "2013-10-19 Post signs around the neighborhood +GarageSale" ]
    assert [t.raw for t in todos.filter_project("+Unpacking")] == [
        "Unpack the guest bedroom +Unpacking due:2013-10-20" ]

def test_todo_highlight(todos):
    todos.parse_raw_entries(["2013-10-25 This is a +Very @cool test"])
    assert todos.todo_items[0].colored == ('plain', ['', ('creation_date', '2013-10-25'), ' This is a ', ('project', '+Very'), ' ', ('context', '@cool'), ' test'])

def test_todos_filter_context_and_project(todos):
    assert [t.raw for t in todos.filter_context_and_project("@phone", "+GarageSale")] == [
        "(B) Schedule Goodwill pickup +GarageSale @phone" ]

def test_todos_update(todos):
    todos.update([
        "(A) 1999-12-24 Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])
    assert [t.raw for t in todos] == [
        "(A) 1999-12-24 Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ]
    todos.update([
        "x 1999-12-25 (A) 1999-12-24 Thank Mom for the dinner @phone"
        "x 1999-11-10 (B) Schedule Goodwill pickup +GarageSale @phone"
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])
    assert [t.raw for t in todos] == [
        "x 1999-12-25 (A) 1999-12-24 Thank Mom for the dinner @phone"
        "x 1999-11-10 (B) Schedule Goodwill pickup +GarageSale @phone"
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ]

def test_todos_complete(todos):
    today = date.today()
    todos.update([
        "(A) 1999-12-24 Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])
    todos[0].complete()
    todos[1].complete()
    assert [t.raw for t in todos] == [
        "x {} (A) 1999-12-24 Thank Mom for the dinner @phone".format(today),
        "x {} (B) Schedule Goodwill pickup +GarageSale @phone".format(today),
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ]
    assert [t.creation_date for t in todos] == [
        "1999-12-24",
        "",
        "",
        "2013-10-19",
        "" ]
    assert [t.completed_date for t in todos] == [
        "{}".format(today),
        "{}".format(today),
        "",
        "",
        "2013-10-01" ]
    assert [t.is_complete() for t in todos] == [ True, True, False, False, True ]
    todos[1].incomplete()
    assert todos[1].raw == "(B) Schedule Goodwill pickup +GarageSale @phone"
    assert todos[1].completed_date == ""

    assert todos[3].creation_date == "2013-10-19"
    assert todos[3].completed_date == ""
    todos[3].complete()
    assert todos[3].creation_date == "2013-10-19"
    assert todos[3].completed_date == "{}".format(today)

def test_todo_incomplete(todos):
    todos.update([
        "x 1999-12-25 (A) 1999-12-24 Thank Mom for the dinner @phone",
        "x 1999-11-10 (B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])
    assert todos[1].creation_date == ""
    assert todos[1].completed_date == "1999-11-10"
    todos[1].incomplete()
    assert todos[1].creation_date == ""
    assert todos[1].completed_date == ""

    assert todos[0].creation_date == "1999-12-24"
    assert todos[0].completed_date == "1999-12-25"
    todos[0].incomplete()
    assert todos[0].creation_date == "1999-12-24"
    assert todos[0].completed_date == ""

def test_todo_is_complete(todos):
    todos.update([
        "x 1999-12-25 (A) 1999-12-24 Thank Mom for the dinner @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x @GroceryStore Eskimo pies" ])
    assert [t.is_complete() for t in todos] == [
        True,
        False,
        False,
        True ]

def test_todo_update(todos):
    t = todos[0]
    t.update("(A) 2000-20-12 Thank Mom for the dinner @phone @home +GiveThanks")
    assert t.raw            == "(A) 2000-20-12 Thank Mom for the dinner @phone @home +GiveThanks"
    assert t.priority       == "A"
    assert t.contexts       == ["@home", "@phone"]
    assert t.projects       == ["+GiveThanks"]
    assert t.creation_date  == "2000-20-12"
    assert t.due_date       == ""
    assert t.completed_date == ""
    assert t.is_complete()  == False

    t = todos[1]
    assert t.raw      == "(B) Schedule Goodwill pickup +GarageSale @phone"
    assert t.contexts == ["@phone"]
    assert t.projects == ["+GarageSale"]
    assert t.priority == "B"

    t = todos[2]
    t.update("x 2013-10-25 Unpack the guest bedroom +Unpacking due:2013-10-20")
    assert t.raw            == "x 2013-10-25 Unpack the guest bedroom +Unpacking due:2013-10-20"
    assert t.priority       == ""
    assert t.contexts       == []
    assert t.projects       == ["+Unpacking"]
    assert t.creation_date  == ""
    assert t.due_date       == "2013-10-20"
    assert t.completed_date == "2013-10-25"
    assert t.is_complete()  == True

    t = todos[3]
    assert t.raw           == "2013-10-19 Post signs around the neighborhood +GarageSale"
    assert t.contexts      == []
    assert t.projects      == ["+GarageSale"]
    assert t.creation_date == "2013-10-19"

    t = todos[4]
    assert t.raw            == "x 2013-10-01 @GroceryStore Eskimo pies"
    assert t.contexts       == ["@GroceryStore"]
    assert t.projects       == []
    assert t.completed_date == "2013-10-01"

def test_todo_add_creation_date(todos, today):
    todos[2].add_creation_date()
    assert todos[2].raw == "{} Unpack the guest bedroom +Unpacking due:2013-10-20".format(today)
    assert todos[2].creation_date == "{}".format(today)

    todos[3].add_creation_date()
    assert todos[3].raw == "2013-10-19 Post signs around the neighborhood +GarageSale"
    assert todos[3].creation_date == "2013-10-19".format(today)

def test_todos_append(todos, today):
    todos.append("THIS IS A TEST @testing")
    todos.append("THIS IS A TEST @testing", add_creation_date=False)
    assert [t.raw for t in todos] == [
        "(A) Thank Mom for the dinner @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies",
        "{} THIS IS A TEST @testing".format(today),
        "THIS IS A TEST @testing".format(today)]
    assert [todo.raw_index for todo in todos.todo_items] == [0, 1, 2, 3, 4, 5, 6]

def test_todos_delete(todos):
    todos.delete(0)
    assert [t.raw for t in todos] == [
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies"]
    assert [todo.raw_index for todo in todos.todo_items] == [0, 1, 2, 3]
    todos.delete(3)
    assert [t.raw for t in todos] == [
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale"]
    assert [todo.raw_index for todo in todos.todo_items] == [0, 1, 2]

def test_todos_insert(todos, today):
    todos.insert(1, "THIS IS A TEST @testing")
    todos.insert(1, "(B) THIS IS ANOTHER TEST @testing")
    todos.insert(1, "(A) THIS IS ANOTHER TEST @testing", add_creation_date=False)
    assert [t.raw for t in todos] == [
        "(A) Thank Mom for the dinner @phone",
        "(A) THIS IS ANOTHER TEST @testing".format(today),
        "(B) {} THIS IS ANOTHER TEST @testing".format(today),
        "{} THIS IS A TEST @testing".format(today),
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies",
    ]
    assert [todo.raw_index for todo in todos.todo_items] == [0, 1, 2, 3, 4, 5, 6, 7]

def test_todos_search(todos):
    assert [t.raw for t in todos.search("the")] == [
        "(A) Thank Mom for the dinner @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale"]
    # one match per line!
    assert [t.search_matches for t in todos.search("the")] == [ ('the',), ('the',), ('the',) ]
    assert [t.raw for t in todos.search("te")] == [
        "(A) Thank Mom for the dinner @phone",
        "Unpack the guest bedroom +Unpacking due:2013-10-20",
        "2013-10-19 Post signs around the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ]
    assert [t.search_matches for t in todos.search("te")] == [ ('the',), ('t be',), ('the',), ('tore',) ]

    assert todos.search(".*") == [ ]
    assert todos.search("{b}") == [ ]

    todos.update([
        "(A) 1999-12-24 .Thank* Mom for the .dinner* @phone",
        "(B) Schedule Goodwill pickup +GarageSale @phone",
        "Unpack the guest {bedroom} +Unpacking due:2013-10-20",
        "2013-10-19 Post signs (around) the neighborhood +GarageSale",
        "x 2013-10-01 @GroceryStore Eskimo pies" ])

    assert [t.raw for t in todos.search(".*")]             == [ "(A) 1999-12-24 .Thank* Mom for the .dinner* @phone" ]
    assert [t.search_matches for t in todos.search(".*")]  == [('.dinner*',)]
    assert [t.raw for t in todos.search("{b}")]            == [ "Unpack the guest {bedroom} +Unpacking due:2013-10-20" ]
    assert [t.search_matches for t in todos.search("{b}")] == [('{bedroom}',)]

def test_todos_swap(todos):
    todos.swap(0, 1)
    assert [todo.raw_index for todo in todos.todo_items] == [1, 0, 2, 3, 4]
    todos.swap(3, 2)
    assert [todo.raw_index for todo in todos.todo_items] == [1, 0, 3, 2, 4]
    todos.swap(0, -1)
    assert [todo.raw_index for todo in todos.todo_items] == [4, 0, 3, 2, 1]
    todos.swap(4, 5)
    assert [todo.raw_index for todo in todos.todo_items] == [1, 0, 3, 2, 4]



#!/usr/bin/env python
# coding=utf-8

"""todotxt-machine

Usage:
  todotxt-machine
  todotxt-machine TODOFILE
  todotxt-machine [--config FILE]
  todotxt-machine (-h | --help)
  todotxt-machine --version

Options:
  -c FILE --config=FILE               Path to your todotxt-machine configuraton file [default: ~/.todotxt-machinerc]
  -h --help                           Show this screen.
  --version                           Show version.
"""

import sys
import os
import random

# import ipdb; # ipdb.set_trace()
# import pprint
# pp = pprint.PrettyPrinter(indent=4).pprint

# Import the correct version of configparser
if sys.version_info[0] >= 3:
    import configparser
    config_parser_module = configparser
elif sys.version_info[0] < 3:
    import ConfigParser
    config_parser_module = ConfigParser

from docopt import docopt

import todotxt_machine
from todotxt_machine.todo import Todos
from todotxt_machine.urwid_ui import UrwidUI
from todotxt_machine.colorscheme import ColorScheme

def exit_with_error(message):
    sys.stderr.write(message.strip(' \n')+'\n')
    print(__doc__.split('\n\n')[1])
    exit(1)

def main():
    random.seed()

    # Parse command line
    arguments = docopt(__doc__, version=todotxt_machine.version)

    # Validate readline editing mode option (docopt doesn't handle this)
    # if arguments['--readline-editing-mode'] not in ['vi', 'emacs']:
    #     exit_with_error("--readline-editing-mode must be set to either vi or emacs\n")

    # Parse config file
    cfg = config_parser_module.ConfigParser(allow_no_value=True)
    cfg.add_section('settings') # make sure we have a setting section so we can call items('settings') and get an empty dict
    cfg.read(os.path.expanduser(arguments['--config']))

    # load the colorscheme defined in the user config, else load the default scheme
    colorscheme = ColorScheme(dict( cfg.items('settings') ).get('colorscheme', 'default'), cfg)

    # Load the todo.txt file specified in the [settings] section of the config file
    # a todo.txt file on the command line takes precedence
    todotxt_file = dict( cfg.items('settings') ).get('file', arguments['TODOFILE'])
    if arguments['TODOFILE']:
        todotxt_file = arguments['TODOFILE']

    if todotxt_file is None:
        exit_with_error("ERROR: No todo file specified. Either specify one as an argument on the command line or set it in your configuration file ({0}).".format(arguments['--config']))

    # expand enviroment variables and username, get canonical path
    todotxt_file_path = os.path.realpath(os.path.expanduser(os.path.expandvars(todotxt_file)))

    # print("Opening: {0}".format(todotxt_file_path))

    if os.path.isdir(todotxt_file_path):
        exit_with_error("ERROR: Specified todo file is a directory.")

    if not os.path.exists(todotxt_file_path):
        directory = os.path.dirname(todotxt_file_path)
        if os.path.isdir(directory):
            # directory exists, but no todo.txt file - create an empty one
            open(todotxt_file_path, 'a').close()
        else:
            exit_with_error("ERROR: The directory: '{0}' does not exist\n\nPlease create the directory or specify a different\ntodo.txt file on the command line.".format(directory))

    try:
        with open(todotxt_file_path, "r") as todotxt_file:
            todos = Todos(todotxt_file.readlines(), todotxt_file_path)
    except:
        exit_with_error("ERROR: unable to open {0}\n\nEither specify one as an argument on the command line or set it in your configuration file ({0}).".format(todotxt_file_path, arguments['--config']))
        todos = Todos([], todotxt_file_path)

    view = UrwidUI(todos, colorscheme)
    view.main()

    # print("Writing: {0}".format(todotxt_file_path))
    view.todos.save()
    exit(0)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

import urwid
import collections

# Modified from http://wiki.goffi.org/wiki/Urwid-satext/en
class AdvancedEdit(urwid.Edit):
    """Edit box with some custom improvments
    new chars:
              - C-a: like 'home'
              - C-e: like 'end'
              - C-k: remove everything on the right of the cursor
              - C-w: remove the word on the back"""

    def __init__(self, parent_ui, *args, **kwargs):
        self.parent_ui = parent_ui
        super(AdvancedEdit, self).__init__(*args, **kwargs)

    def setCompletionMethod(self, callback):
        """Define method called when completion is asked
        @callback: method with 2 arguments:
                    - the text to complete
                    - if there was already a completion, a dict with
                        - 'completed':last completion
                        - 'completion_pos': cursor position where the completion starts
                        - 'position': last completion cursor position
                      this dict must be used (and can be filled) to find next completion)
                   and which return the full text completed"""
        self.completion_cb = callback
        self.completion_data = {}

    def keypress(self, size, key):
        # import ipdb; ipdb.set_trace()
        if key == 'ctrl a':
            key = 'home'
        elif key == 'ctrl e':
            key = 'end'
        elif key == 'ctrl k':
            self.parent_ui.yanked_text = self.edit_text[self.edit_pos:]
            self._delete_highlighted()
            self.set_edit_text(self.edit_text[:self.edit_pos])
        elif key == 'ctrl y':
            self.set_edit_text(
                self.edit_text[:self.edit_pos] +
                self.parent_ui.yanked_text +
                self.edit_text[self.edit_pos:])
            self.set_edit_pos(self.edit_pos + len(self.parent_ui.yanked_text))
        elif key == 'ctrl w':
            before = self.edit_text[:self.edit_pos]
            pos = before.rstrip().rfind(" ")+1
            self.parent_ui.yanked_text = self.edit_text[pos:self.edit_pos]
            self.set_edit_text(before[:pos] + self.edit_text[self.edit_pos:])
            self.set_edit_pos(pos)
        elif key in ['meta b', 'ctrl b']:
            before = self.edit_text[:self.edit_pos]
            pos = before.rstrip().rfind(" ")+1
            self.set_edit_pos(pos)
        elif key in ['meta f', 'ctrl f']:
            after = self.edit_text[self.edit_pos:]
            pos = after.rstrip().find(" ")+1
            self.set_edit_pos(self.edit_pos+pos)
        elif key == 'tab':
            try:
                before = self.edit_text[:self.edit_pos]
                if self.completion_data:
                    if (not self.completion_data['completed']
                        or self.completion_data['position'] != self.edit_pos
                        or not before.endswith(self.completion_data['completed'])):
                        self.completion_data.clear()
                    else:
                        before = before[:-len(self.completion_data['completed'])]
                complet = self.completion_cb(before, self.completion_data)
                self.completion_data['completed'] = complet[len(before):]
                self.set_edit_text(complet+self.edit_text[self.edit_pos:])
                self.set_edit_pos(len(complet))
                self.completion_data['position'] = self.edit_pos
                return
            except AttributeError:
                #No completion method defined
                pass
        return super(AdvancedEdit, self).keypress(size, key)

class SearchWidget(urwid.Edit):
    def __init__(self, parent_ui, edit_text=""):
        self.parent_ui = parent_ui
        super(SearchWidget, self).__init__(edit_text=edit_text)

    def keypress(self, size, key):
        if key == 'enter':
            self.parent_ui.finalize_search()
        return super(SearchWidget, self).keypress(size, key)

class TodoWidget(urwid.Button):
    def __init__(self, todo, colorscheme, parent_ui, editing=False, wrapping='clip', border='no border'):
        super(TodoWidget, self).__init__("")
        self.todo        = todo
        self.wrapping    = wrapping
        self.border      = border
        self.colorscheme = colorscheme
        self.parent_ui   = parent_ui
        self.editing     = editing
        # urwid.connect_signal(self, 'click', callback)
        if editing:
            self.edit_item()
        else:
            self.update_todo()

    def selectable(self):
        return True

    def update_todo(self):
        if self.parent_ui.searching and self.parent_ui.search_string:
            text = urwid.Text(self.todo.highlight_search_matches(), wrap=self.wrapping)
        else:
            if self.border == 'bordered':
                text = urwid.Text(self.todo.highlight(show_due_date=False, show_contexts=False, show_projects=False), wrap=self.wrapping)
            else:
                text = urwid.Text(self.todo.colored, wrap=self.wrapping)

        if self.border == 'bordered':
            lt=''
            if self.todo.due_date:
                lt = ('due_date', "due:{0}".format(self.todo.due_date))
            t = []
            t.append( ('context', ' '.join(self.todo.contexts)) )
            if self.todo.contexts and self.todo.projects:
                t.append(' ')
            t.append( ('project', ' '.join(self.todo.projects)) )
            bc = 'plain'
            if self.todo.priority and self.todo.priority in "ABCDEF":
                bc = "priority_{0}".format(self.todo.priority.lower())
            text = TodoLineBox(text, top_left_title=lt, bottom_right_title=t, border_color=bc, )
        self._w = urwid.AttrMap( urwid.AttrMap(
            text,
            None, 'selected'),
            None, self.colorscheme.focus_map)

    def edit_item(self):
        self.editing = True
        self.edit_widget = AdvancedEdit(self.parent_ui, caption="", edit_text=self.todo.raw)
        self.edit_widget.setCompletionMethod(self.completions)
        self._w = urwid.AttrMap(self.edit_widget, 'plain_selected')

    def completions(self, text, completion_data={}):
        space = text.rfind(" ")
        start = text[space+1:]
        words = self.parent_ui.todos.all_contexts() + self.parent_ui.todos.all_projects()
        try:
            start_idx=words.index(completion_data['last_word'])+1
            if start_idx == len(words):
                start_idx = 0
        except (KeyError,ValueError):
            start_idx = 0
        for idx in list(range(start_idx,len(words))) + list(range(0,start_idx)):
            if words[idx].lower().startswith(start.lower()):
                completion_data['last_word'] = words[idx]
                return text[:space+1] + words[idx] + (': ' if space < 0 else '')
        return text

    def save_item(self):
        self.todo.update(self._w.original_widget.edit_text.strip())
        self.update_todo()
        if self.parent_ui.filter_panel_is_open:
            self.parent_ui.update_filters(new_contexts=self.todo.contexts, new_projects=self.todo.projects)
        self.editing = False

    def keypress(self, size, key):
        if self.editing:
            if key in ['down', 'up']:
                return None # don't pass up or down to the ListBox
            elif key is 'enter':
                self.save_item()
                return key
            else:
                return self._w.keypress(size, key)
        else:
            if key in ['enter', 'e', 'A']:
                self.edit_item()
                return key
            else:
                return key

class TodoLineBox(urwid.WidgetDecoration, urwid.WidgetWrap):

    def __init__(self, original_widget, top_left_title="", bottom_right_title="", border_color='plain',
                 tlcorner=u'┌', tline=u'─', lline=u'│',
                 trcorner=u'┐', blcorner=u'└', rline=u'│',
                 bline=u'─', brcorner=u'┘'):
        """
        Draw a line around original_widget.

        Use 'title' to set an initial title text with will be centered
        on top of the box.

        You can also override the widgets used for the lines/corners:
            tline: top line
            bline: bottom line
            lline: left line
            rline: right line
            tlcorner: top left corner
            trcorner: top right corner
            blcorner: bottom left corner
            brcorner: bottom right corner

        """

        tline, bline = urwid.AttrMap(urwid.Divider(tline), border_color),   urwid.AttrMap(urwid.Divider(bline), border_color)
        lline, rline = urwid.AttrMap(urwid.SolidFill(lline), border_color), urwid.AttrMap(urwid.SolidFill(rline), border_color)
        tlcorner, trcorner = urwid.AttrMap(urwid.Text(tlcorner), border_color), urwid.AttrMap(urwid.Text(trcorner), border_color)
        blcorner, brcorner = urwid.AttrMap(urwid.Text(blcorner), border_color), urwid.AttrMap(urwid.Text(brcorner), border_color)

        self.ttitle_widget = urwid.Text(top_left_title)
        self.tline_widget = urwid.Columns([
            ('fixed', 1, tline),
            ('flow', self.ttitle_widget),
            tline,
        ])
        self.btitle_widget = urwid.Text(bottom_right_title)
        self.bline_widget = urwid.Columns([
            bline,
            ('flow', self.btitle_widget),
            ('fixed', 1, bline),
        ])

        middle = urwid.Columns([
            ('fixed', 1, lline),
            original_widget,
            ('fixed', 1, rline),
        ], box_columns=[0, 2], focus_column=1)

        top = urwid.Columns([
            ('fixed', 1, tlcorner),
            self.tline_widget,
            ('fixed', 1, trcorner)
        ])

        bottom = urwid.Columns([
            ('fixed', 1, blcorner),
            self.bline_widget,
            ('fixed', 1, brcorner)
        ])

        pile = urwid.Pile([('flow', top), middle, ('flow', bottom)], focus_item=1)

        urwid.WidgetDecoration.__init__(self, original_widget)
        urwid.WidgetWrap.__init__(self, pile)


class ViPile(urwid.Pile):
    def __init__(self, widget_list, focus_item=None):
        """Pile with Vi-like navigation."""
        super(ViPile, self).__init__(widget_list, focus_item)
        command_map = urwid.command_map.copy()
        command_map['j'] = urwid.CURSOR_DOWN
        command_map['k'] = urwid.CURSOR_UP
        self._command_map = command_map


class ViColumns(urwid.Columns):
    def __init__(self, widget_list, dividechars=0, focus_column=None, min_width=1, box_columns=None):
        super(ViColumns, self).__init__(widget_list, dividechars, focus_column, min_width, box_columns)
        command_map = urwid.command_map.copy()
        command_map['L'] = urwid.CURSOR_RIGHT
        command_map['H'] = urwid.CURSOR_LEFT
        self._command_map = command_map

class ViListBox(urwid.ListBox):
    def __init__(self, *args, **kwargs):
        super(ViListBox, self).__init__(*args, **kwargs)
        command_map = urwid.command_map.copy()
        command_map['j'] = urwid.CURSOR_DOWN
        command_map['k'] = urwid.CURSOR_UP
        self._command_map = command_map

class UrwidUI:
    def __init__(self, todos, colorscheme):
        self.wrapping    = collections.deque(['clip', 'space'])
        self.border      = collections.deque(['no border', 'bordered'])
        self.sorting     = collections.deque(["Unsorted", "Descending", "Ascending"])
        self.sorting_display = {"Unsorted": "-", "Descending": "v", "Ascending": "^"}

        self.todos       = todos

        self.colorscheme = colorscheme
        self.palette     = [ (key, '', '', '', value['fg'], value['bg']) for key, value in self.colorscheme.colors.items() ]

        self.active_projects = []
        self.active_contexts = []

        self.toolbar_is_open       = False
        self.help_panel_is_open    = False
        self.filter_panel_is_open  = False
        self.filtering             = False
        self.searching             = False
        self.search_string         = ''
        self.yanked_text           = ''

    def visible_lines(self):
        lines = self.loop.screen_size[1] - 1 # minus one for the header
        if self.toolbar_is_open:
            lines -= 1
        if self.searching:
            lines -= 1
        return lines

    def move_selection_down(self):
        self.listbox.keypress((0, self.visible_lines()), 'down')

    def move_selection_up(self):
        self.listbox.keypress((0, self.visible_lines()), 'up')

    def move_selection_top(self):
        self.listbox.set_focus(0)

    def move_selection_bottom(self):
        self.listbox.set_focus(len(self.listbox.body)-1)

    def toggle_help_panel(self, button=None):
        if self.filter_panel_is_open:
            self.toggle_filter_panel()
        if self.help_panel_is_open:
            self.view.contents.pop()
            self.help_panel_is_open = False
            # set header line to word-wrap contents
            # for header_column in self.frame.header.original_widget.contents:
            #     header_column[0].set_wrap_mode('space')
        else:
            self.help_panel = self.create_help_panel()
            self.view.contents.append( (self.help_panel, self.view.options(width_type='weight', width_amount=3)) )
            self.view.set_focus(1)
            self.help_panel_is_open = True
            # set header line to clip contents
            # for header_column in self.frame.header.original_widget.contents:
            #     header_column[0].set_wrap_mode('clip')

    def toggle_sorting(self, button=None):
        self.delete_todo_widgets()
        self.sorting.rotate(1)
        if self.sorting[0] == 'Ascending':
            self.todos.sorted()
        elif self.sorting[0] == 'Descending':
            self.todos.sorted_reverse()
        elif self.sorting[0] == 'Unsorted':
            self.todos.sorted_raw()
        self.reload_todos_from_memory()
        self.move_selection_top()
        self.update_header()

    def toggle_filter_panel(self, button=None):
        if self.help_panel_is_open:
            self.toggle_help_panel()
        if self.filter_panel_is_open:
            self.view.contents.pop()
            self.filter_panel_is_open = False
        else:
            self.filter_panel = self.create_filter_panel()
            self.view.contents.append( (self.filter_panel, self.view.options(width_type='weight', width_amount=1)) )
            self.filter_panel_is_open = True

    def toggle_wrapping(self, checkbox=None, state=None):
        self.wrapping.rotate(1)
        for widget in self.listbox.body:
            widget.wrapping = self.wrapping[0]
            widget.update_todo()
        if self.toolbar_is_open:
            self.update_header()

    def toggle_border(self, checkbox=None, state=None):
        self.border.rotate(1)
        for widget in self.listbox.body:
            widget.border = self.border[0]
            widget.update_todo()
        if self.toolbar_is_open:
            self.update_header()

    def toggle_toolbar(self):
        self.toolbar_is_open = not self.toolbar_is_open
        self.update_header()

    def swap_down(self):
        focus, focus_index = self.listbox.get_focus()
        if not self.filtering and not self.searching:
            if focus_index+1 < len(self.listbox.body):
                self.todos.swap(focus_index, focus_index + 1)
                self.listbox.body[focus_index].todo = self.todos[focus_index]
                self.listbox.body[focus_index+1].todo = self.todos[focus_index+1]
                self.listbox.body[focus_index].update_todo()
                self.listbox.body[focus_index+1].update_todo()
                self.move_selection_down()

    def swap_up(self):
        focus, focus_index = self.listbox.get_focus()
        if not self.filtering and not self.searching:
            if focus_index > 0:
                self.todos.swap(focus_index, focus_index - 1)
                self.listbox.body[focus_index].todo = self.todos[focus_index]
                self.listbox.body[focus_index-1].todo = self.todos[focus_index-1]
                self.listbox.body[focus_index].update_todo()
                self.listbox.body[focus_index-1].update_todo()
                self.move_selection_up()

    def save_todos(self, button=None):
        self.todos.save()
        self.update_header("Saved")

    def reload_todos_from_file(self, button=None):
        self.delete_todo_widgets()

        self.todos.reload_from_file()

        for t in self.todos.todo_items:
            self.listbox.body.append( TodoWidget(t, self.colorscheme, self, wrapping=self.wrapping[0], border=self.border[0]) )

        self.update_header("Reloaded")

    def keystroke(self, input):
        focus, focus_index = self.listbox.get_focus()

        if input in ('q', 'Q'):
            raise urwid.ExitMainLoop()

        # Movement
        elif input is 'g':
            self.move_selection_top()
        elif input is 'G':
            self.move_selection_bottom()
        elif input is 'J':
            self.swap_down()
        elif input is 'K':
            self.swap_up()
        elif input is 'tab':
            current_focus = self.frame.get_focus()
            if current_focus == 'body':

                if self.filter_panel_is_open and self.toolbar_is_open:

                    if self.view.focus_position == 1:
                        self.view.focus_position = 0
                        self.frame.focus_position = 'header'
                    elif self.view.focus_position == 0:
                        self.view.focus_position = 1

                elif self.toolbar_is_open:
                    self.frame.focus_position = 'header'

                elif self.filter_panel_is_open:
                    if self.view.focus_position == 1:
                        self.view.focus_position = 0
                    elif self.view.focus_position == 0:
                        self.view.focus_position = 1

            elif current_focus == 'header':
                self.frame.focus_position = 'body'

        # View options
        elif input in ['h', '?']:
            self.toggle_help_panel()
        elif input is 't':
            self.toggle_toolbar()
        elif input is 'f':
            self.toggle_filter_panel()
        elif input is 'F':
            self.clear_filters()
        elif input is 'w':
            self.toggle_wrapping()
        elif input is 'b':
            self.toggle_border()
        elif input is 's':
            self.toggle_sorting()

        elif input is '/':
            self.start_search()
        elif input is 'C':
            if self.searching:
                self.clear_search_term()

        # Editing
        elif input is 'x':
            if focus.todo.is_complete():
                focus.todo.incomplete()
            else:
                focus.todo.complete()
            focus.update_todo()
            self.update_header()
        elif input is 'D':
            if self.todos.todo_items:
                i = focus.todo.raw_index
                self.todos.delete(i)
                del self.listbox.body[focus_index]
                self.update_header()

        elif input is 'n':
            self.add_new_todo(position='append')
        elif input is 'O':
            self.add_new_todo(position='insert_before')
        elif input is 'o':
            self.add_new_todo(position='insert_after')

        # Save current file
        elif input is 'S':
            self.save_todos()

        # Reload original file
        elif input is 'R':
            self.reload_todos_from_file()

    def add_new_todo(self, position=False):
        focus_index = self.listbox.get_focus()[1]
        if self.filtering:
            position = 'append'

        if position is 'append':
            new_index = self.todos.append('', add_creation_date=False)
            self.listbox.body.append(TodoWidget(self.todos[new_index], self.colorscheme, self, editing=True, wrapping=self.wrapping[0], border=self.border[0]))
        else:
            if position is 'insert_after':
                new_index = self.todos.insert(focus_index+1, '', add_creation_date=False)
            elif position is 'insert_before':
                new_index = self.todos.insert(focus_index, '', add_creation_date=False)

            self.listbox.body.insert(new_index, TodoWidget(self.todos[new_index], self.colorscheme, self, editing=True, wrapping=self.wrapping[0], border=self.border[0]))

        if position:
            if self.filtering:
                self.listbox.set_focus(len(self.listbox.body)-1)
            else:
                self.listbox.set_focus(new_index)
            # edit_widget = self.listbox.body[new_index]._w
            # edit_widget.edit_text += ' '
            # edit_widget.set_edit_pos(len(self.todos[new_index].raw) + 1)
            self.update_header()

    def create_header(self, message=""):
        return urwid.AttrMap(
            urwid.Columns( [
                urwid.Text( [
                    ('header_todo_count', "{0} Todos ".format(self.todos.__len__())),
                    ('header_todo_pending_count', " {0} Pending ".format(self.todos.pending_items_count())),
                    ('header_todo_done_count', " {0} Done ".format(self.todos.done_items_count())),
                ]),
                # urwid.Text( " todotxt-machine ", align='center' ),
                urwid.Text( ('header_file', "{0}  {1} ".format(message, self.todos.file_path)), align='right' )
            ]), 'header')

    def create_toolbar(self):
        return urwid.AttrMap(urwid.Columns( [
            urwid.Padding(
            urwid.AttrMap(
            urwid.CheckBox([('header_file', 'w'), 'ord wrap'], state=(self.wrapping[0] == 'space'), on_state_change=self.toggle_wrapping),
            'header', 'plain_selected'), right=2 ),

            urwid.Padding(
            urwid.AttrMap(
            urwid.CheckBox([('header_file', 'b'), 'orders'], state=(self.border[0] == 'bordered'), on_state_change=self.toggle_border),
            'header', 'plain_selected'), right=2 ),

            urwid.Padding(
            urwid.AttrMap(
            urwid.Button([('header_file', 'R'), 'eload'], on_press=self.reload_todos_from_file),
            'header', 'plain_selected'), right=2 ),

            urwid.Padding(
            urwid.AttrMap(
            urwid.Button([('header_file', 'S'), 'ave'], on_press=self.save_todos),
            'header', 'plain_selected'), right=2 ),

            urwid.Padding(
            urwid.AttrMap(
            urwid.Button([('header_file', 's'), 'ort: '+self.sorting_display[self.sorting[0]]], on_press=self.toggle_sorting),
            'header', 'plain_selected'), right=2 ),

            urwid.Padding(
            urwid.AttrMap(
            urwid.Button([('header_file', 'f'), 'ilter'], on_press=self.toggle_filter_panel),
            'header', 'plain_selected'), right=2 )
        ] ), 'header')

    def search_box_updated(self, edit_widget, new_contents):
        old_contents = edit_widget.edit_text
        self.search_string = new_contents
        self.search_todo_list(self.search_string)

    def search_todo_list(self, search_string=""):
        if search_string:
            self.delete_todo_widgets()

            self.searching = True

            for t in self.todos.search(search_string):
                self.listbox.body.append( TodoWidget(t, self.colorscheme, self, wrapping=self.wrapping[0], border=self.border[0]) )

    def start_search(self):
        self.searching = True
        self.update_footer()
        self.frame.set_focus('footer')

    def finalize_search(self):
        self.search_string = ''
        self.frame.set_focus('body')
        for widget in self.listbox.body:
            widget.update_todo()

    def clear_search_term(self, button=None):
        self.delete_todo_widgets()
        self.searching = False
        self.search_string = ''
        self.update_footer()
        self.reload_todos_from_memory()

    def create_footer(self):
        if self.searching:
            self.search_box = SearchWidget(self, edit_text=self.search_string)
            w = urwid.AttrMap(urwid.Columns([
                (1, urwid.Text('/')),
                self.search_box,
                (16, urwid.AttrMap(
                    urwid.Button([('header_file', 'C'), 'lear Search'], on_press=self.clear_search_term),
                    'header', 'plain_selected') )
            ]), 'footer')
            urwid.connect_signal(self.search_box, 'change', self.search_box_updated)
        else:
            w = None
        return w

    def create_help_panel(self):
        header_highlight = 'plain_selected'
        return urwid.AttrMap(
            urwid.LineBox(
            urwid.Padding(
            ViListBox(
                [ urwid.Divider() ] +

                [ urwid.AttrWrap(urwid.Text("""
General
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
h, ?         - show / hide this help message
q            - quit and save
t            - show / hide toolbar
w            - toggle word wrap
b            - toggle borders on todo items
S            - save current todo file
R            - reload the todo file (discarding changes)
""")] +

                [ urwid.AttrWrap(urwid.Text("""
Movement
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
mouse click  - select any todo, checkbox or button
j, down      - move selection down
k, up        - move selection up
g, page up   - move selection to the top item
G, page down - move selection to the bottom item
left, right  - move selection between todos and filter panel
H, L
tab          - toggle focus between todos, filter panel, and toolbar
""")] +

                [ urwid.AttrWrap(urwid.Text("""
Manipulating Todo Items
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
x            - complete / un-complete selected todo item
n            - add a new todo to the end of the list
o            - add a todo after the selected todo (when not filtering)
O            - add a todo before the selected todo (when not filtering)
enter, A, e  - edit the selected todo
D            - delete the selected todo
J            - swap with item below
K            - swap with item above
""")] +

                [ urwid.AttrWrap(urwid.Text("""
While Editing a Todo
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
tab          - tab complete contexts and projects
return       - save todo item
left, right  - move cursor left and right
ctrl-b       - move cursor backwards (left) by one word
ctrl-f       - move cursor forwards (right) by one word
home, end    - move cursor the beginning or end of the line
ctrl-a, ctrl-e
ctrl-w       - delete one word backwards
ctrl-k       - delete from the cursor to the end of the line
ctrl-y       - paste last deleted text
""")] +

                [ urwid.AttrWrap(urwid.Text("""
Sorting
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
s            - toggle sort order (Unsorted, Ascending, Descending)
               sort order is saved on quit
""")] +
                [ urwid.AttrWrap(urwid.Text("""
Filtering
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
f            - open / close the filtering panel
F            - clear any active filters
""")] +
                [ urwid.AttrWrap(urwid.Text("""
Searching
""".strip()), header_highlight) ] +
                # [ urwid.Divider(u'─') ] +

                [ urwid.Text("""
/            - start search
enter        - finalize search
C            - clear search
""")]
            ),
            left=1, right=1, min_width=10 ), title='Key Bindings'), 'default')

    def create_filter_panel(self):
        w = urwid.AttrMap(
            urwid.Padding(
            urwid.ListBox(
                [
                    ViPile(
                        [ urwid.Text('Contexts & Projects', align='center') ] +
                        [ urwid.Divider(u'─') ] +
                        [urwid.AttrWrap(urwid.CheckBox(c, state=(c in self.active_contexts), on_state_change=self.checkbox_clicked, user_data=['context', c]), 'context_dialog_color', 'context_selected') for c in self.todos.all_contexts()] +
                        [ urwid.Divider(u'─') ] +
                        [urwid.AttrWrap(urwid.CheckBox(p, state=(p in self.active_projects), on_state_change=self.checkbox_clicked, user_data=['project', p]), 'project_dialog_color', 'project_selected') for p in self.todos.all_projects()] +
                        [ urwid.Divider(u'─') ] +
                        [ urwid.AttrMap(urwid.Button(['Clear ', ('header_file_dialog_color','F'), 'ilters'], on_press=self.clear_filters), 'dialog_color', 'plain_selected') ]
                    )
                ] +
                [ urwid.Divider() ],
            ),
            left=1, right=1, min_width=10 )
            ,
        'dialog_color')

        bg = urwid.AttrWrap(urwid.SolidFill(u" "), 'dialog_background') # u"\u2592"
        shadow = urwid.AttrWrap(urwid.SolidFill(u" "), 'dialog_shadow')

        bg = urwid.Overlay( shadow, bg,
            ('fixed left', 2), ('fixed right', 1),
            ('fixed top', 2), ('fixed bottom', 1))
        w = urwid.Overlay( w, bg,
            ('fixed left', 1), ('fixed right', 2),
            ('fixed top', 1), ('fixed bottom', 2))
        return w

    def delete_todo_widgets(self):
        for i in range(len(self.listbox.body)-1, -1, -1):
            self.listbox.body.pop(i)

    def reload_todos_from_memory(self):
        for t in self.todos.todo_items:
            self.listbox.body.append( TodoWidget(t, self.colorscheme, self, wrapping=self.wrapping[0], border=self.border[0]) )

    def clear_filters(self, button=None):
        self.delete_todo_widgets()
        self.reload_todos_from_memory()

        self.active_projects = []
        self.active_contexts = []
        self.filtering = False
        self.view.set_focus(0)
        self.update_filters()

    def checkbox_clicked(self, checkbox, state, data):
        if state:
            if data[0] == 'context':
                self.active_contexts.append(data[1])
            else:
                self.active_projects.append(data[1])
        else:
            if data[0] == 'context':
                self.active_contexts.remove(data[1])
            else:
                self.active_projects.remove(data[1])

        if self.active_projects or self.active_contexts:
            self.filter_todo_list()
            self.view.set_focus(0)
        else:
            self.clear_filters()

    def filter_todo_list(self):
        self.delete_todo_widgets()

        for t in self.todos.filter_contexts_and_projects(self.active_contexts, self.active_projects):
            self.listbox.body.append( TodoWidget(t, self.colorscheme, self, wrapping=self.wrapping[0], border=self.border[0]) )

        self.filtering = True

    def update_filters(self, new_contexts=[], new_projects=[]):
        if self.active_contexts:
            for c in new_contexts:
                self.active_contexts.append(c)
        if self.active_projects:
            for p in new_projects:
                self.active_projects.append(p)
        self.update_filter_panel()

    def update_filter_panel(self):
        self.filter_panel = self.create_filter_panel()
        if len(self.view.widget_list) > 1:
            self.view.widget_list.pop()
            self.view.widget_list.append(self.filter_panel)

    def update_header(self, message=""):
        if self.toolbar_is_open:
            self.frame.header = urwid.Pile([self.create_header(message), self.create_toolbar()])
        else:
            self.frame.header = self.create_header(message)

    def update_footer(self, message=""):
        self.frame.footer = self.create_footer()

    def main(self):
        self.header = self.create_header()
        self.footer = self.create_footer()

        self.listbox = ViListBox(urwid.SimpleListWalker(
            [TodoWidget(t, self.colorscheme, self) for t in self.todos.todo_items]
        ))

        self.frame  = urwid.Frame(urwid.AttrMap(self.listbox, 'plain'), header=self.header, footer=self.footer)

        self.view = ViColumns([
            ('weight', 2, self.frame )
         ])

        self.loop = urwid.MainLoop(self.view, self.palette, unhandled_input=self.keystroke)
        self.loop.screen.set_terminal_properties(colors=256)
        self.loop.run()

